# -*- coding: utf-8 -*-

"""
This module contains the Genens estimators.
"""
import os
from typing import Union

from genens.base import GenensBase
from genens.config.config_templates import clf_default_config, default_raw_config
from genens.config.genens_config import GenensConfig, parse_config
from genens.workflow.evaluate import EvaluatorBase

from sklearn.base import ClassifierMixin, RegressorMixin


class GenensClassifier(GenensBase, ClassifierMixin):
    def __init__(self, config: Union[GenensConfig, str] = None, use_base_config: bool = True, evo_kwargs: dict = None,
                 n_jobs: int = 1, scorer=None, pop_size: int = 100, n_gen: int = 15,
                 timeout: int = None, evaluator: EvaluatorBase = None,
                 max_evo_seconds: int = None, **kwargs):

        if config is None:
            config = clf_default_config()

        if isinstance(config, str):
            base_config = default_raw_config() if use_base_config else None
            config = parse_config(config, base_config=base_config, evo_kwargs=evo_kwargs)

        super().__init__(config,
                         n_jobs=n_jobs,
                         scorer=scorer,
                         pop_size=pop_size,
                         n_gen=n_gen,
                         timeout=timeout,
                         evaluator=evaluator,
                         max_evo_seconds=max_evo_seconds,
                         **kwargs)


class GenensRegressor(GenensBase, RegressorMixin):
    def __init__(self, config: Union[GenensConfig, str] = None, use_base_config: bool = True, evo_kwargs: dict = None,
                 n_jobs: int = 1, scorer=None, pop_size: int = 100, n_gen: int = 15,
                 timeout: int = None, evaluator: EvaluatorBase = None,
                 max_evo_seconds: int = None, **kwargs):
        if config is None:
            raise ValueError("Regressor default config is not available yet.")

        if isinstance(config, str):
            base_config = default_raw_config() if use_base_config else None
            config = parse_config(config, base_config=base_config, evo_kwargs=evo_kwargs)

        super().__init__(config,
                         n_jobs=n_jobs,
                         scorer=scorer,
                         pop_size=pop_size,
                         n_gen=n_gen,
                         timeout=timeout,
                         evaluator=evaluator,
                         max_evo_seconds=max_evo_seconds,
                         **kwargs)



# -*- coding: utf-8 -*-
"""
Module for visualization of the trees.
"""

import pygraphviz as pgv


def create_graph(gp_tree, file_name):
    graph = pgv.AGraph()
    i = 0

    def add_node(node, child_list):
        nonlocal i
        node_id = i
        i = i + 1

        graph.add_node(node_id, label=node.name)

        for val in child_list:
            graph.add_edge(node_id, val)

        return node_id

    gp_tree.run_tree(add_node)

    graph.layout(prog='dot')
    graph.draw(file_name, prog='dot')
    graph.close()





# -*- coding: utf-8 -*-
import os

from sklearn import feature_selection
from genens.config.genens_config import GenensConfig, parse_config

file_dir = os.path.dirname(os.path.abspath(__file__))
BASE_CONFIG_PATH = file_dir + '/config_default.yaml'
BASE_CLF_PATH = file_dir + '/config_clf.yaml'


def default_raw_config() -> GenensConfig:
    return parse_config(BASE_CONFIG_PATH, evo_kwargs=default_kwargs())


def clf_default_config() -> GenensConfig:
    default_config = default_raw_config()
    return parse_config(BASE_CLF_PATH, base_config=default_config, evo_kwargs=clf_kwargs())


def default_kwargs() -> dict:
    return {k: {} for k in ['cPred', 'cPipe', 'cData', 'cFeatSelect', 'cScale']}


def clf_kwargs() -> dict:
    """
    Creates Genens hyperparameter configuration for classifier problems.

    Returns:
        dict: Hyperparameter configuration for classifier problems.
    """

    ensemble_kwargs = {
        "voting": {
            "voting": ["hard", "soft"]
        },
        "stacking": {
            "cv": [None, 3, 7],
            "passthrough": [True, False]
        }
    }

    clf_kwargs = {
        "ada": {
            "n_estimators": [5, 10, 50, 100, 200],
            "algorithm": ["SAMME", "SAMME.R"],
            "learning_rate": [0.001, 0.01, 0.1, 0.5, 1.0]
        },
        "bagging": {
            "n_estimators": [5, 10, 50, 100, 200],
            "oob_score": [True, False],
            "max_samples": [0.2, 0.5, 0.8, 1.0],
            "max_features": [0.2, 0.5, 0.8, 1.0]
        },
        "KNeighbors": {
            "n_neighbors": [1, 2, 5],
            "algorithm": ["auto", "ball_tree", "kd_tree", "brute"]
        },
        "LinearSVC": {
            "loss": ["hinge", "squared_hinge"],
            "penalty": ["l1", "l2"],
            "C": [0.1, 0.5, 1.0, 2, 5, 10, 15],
            "tol": [0.0001, 0.001, 0.01]
        },
        "SVC": {
            "C": [0.1, 0.5, 1.0, 2, 5, 10, 15],
            "gamma": ["scale", 0.0001, 0.001, 0.01, 0.1, 0.5],
            "tol": [0.0001, 0.001, 0.01],
            "probability": [True, False]
        },
        "logR": {
            "penalty": ["l1", "l2"],
            "C": [0.1, 0.5, 1.0, 2, 5, 10, 15],
            "tol": [0.0001, 0.001, 0.01],
            "solver": ["newton-cg", "lbfgs", "liblinear", "sag", "saga"]
        },
        "Perceptron": {
            "penalty": ["None", "l2", "l1", "elasticnet"],
            "max_iter": [1, 2, 5, 10, 100],
            "alpha": [0.0001, 0.001, 0.01]
        },
        "SGD": {
            "penalty": ["none", "l2", "l1", "elasticnet"],
            "loss": ["hinge", "log", "modified_huber", "squared_hinge", "perceptron"],
            "max_iter": [10, 100, 200],
            "tol": [0.0001, 0.001, 0.01],
            "alpha": [0.0001, 0.001, 0.01],
            "l1_ratio": [0, 0.15, 0.5, 1],
            "epsilon": [0.01, 0.05, 0.1, 0.5],
            "learning_rate": ["constant", "optimal"],
            "eta0": [0.01, 0.1, 0.5],  # wild guess
            "power_t": [0.1, 0.5, 1, 2]  # dtto
        },
        "PAC": {
            "loss": ["hinge", "squared_hinge"],
            "C": [0.1, 0.5, 1.0, 2, 5, 10, 15]
        },
        "LDA": {
            "solver": ["lsqr", "eigen"],
            "shrinkage": [None, "auto", 0.1, 0.5, 1.0]
        },
        "QDA": {
            "reg_param": [0.0, 0.1, 0.5, 1],
            "tol": [0.0001, 0.001, 0.01]
        },
        "MLP": {
            "activation": ["identity", "logistic", "relu"],
            "solver": ["lbfgs", "sgd", "adam"],
            "alpha": [0.0001, 0.001, 0.01],
            "learning_rate": ["constant", "invscaling", "adaptive"],
            "tol": [0.0001, 0.001, 0.01],
            "max_iter": [10, 100, 200],
            "learning_rate_init": [0.0001, 0.001, 0.01],
            "power_t": [0.1, 0.5, 1, 2],
            "momentum": [0.1, 0.5, 0.9],
            "hidden_layer_sizes": [(100,), (50,), (20,), (10,)]
        },
        "DT": {
            "criterion": ["gini", "entropy"],
            "max_features": [0.05, 0.1, 0.25, 0.5, 0.75, 1],
            "max_depth": [1, 2, 5, 10, 15, 25, 50, 100],
            "min_samples_split": [2, 5, 10, 20],
            "min_samples_leaf": [1, 2, 5, 10, 20]
        },
        "gaussianNB": {},
        "gradBoosting": {
            "loss": ["deviance", "exponential"],
            "learning_rate": [0.0001, 0.001, 0.01, 0.1, 0.2, 0.3],
            "n_estimators": [20, 50, 100],
            "subsample": [0.3, 0.5, 0.75, 1.0],
            "max_depth": [None, 3, 5, 10, 20, 50, 100],
            "min_samples_split": [2, 5, 10, 15, 20],
            "min_samples_leaf": [1, 2, 5, 10, 20],
        },
        "randomForest": {
            "n_estimators": [10, 50, 100, 150],
            "criterion": ["gini", "entropy"],
            "max_features": [None, "sqrt", 0.05, 0.2, 0.5, 0.8],
            "max_depth": [None, 3, 5, 10, 20, 50, 100],
            "min_samples_split": [2, 5, 10, 15, 20],
            "min_samples_leaf": [1, 2, 5, 10, 20],
            "bootstrap": [True, False],
            "ccp_alpha": [0.0, 0.001, 0.01, 0.03, 0.05],
            "oob_score": [True, False]
        },
        "extraTrees": {
            "n_estimators": [10, 50, 100, 150],
            "criterion": ["gini", "entropy"],
            "max_features": [None, "sqrt", 0.05, 0.2, 0.5, 0.8],
            "max_depth": [None, 3, 5, 10, 20, 50, 100],
            "min_samples_split": [2, 5, 10, 15, 20],
            "min_samples_leaf": [1, 2, 5, 10, 20],
            "bootstrap": [True, False],
            "ccp_alpha": [0.0, 0.001, 0.01, 0.03, 0.05],
            "oob_score": [True, False]
        }
    }

    transform_kwargs = {
        "NMF": {
            "feat_frac": [0.1, 0.5, 0.8, None],
            "solver": ["cd", "mu"],
            "tol": [0.0001, 0.001, 0.01]
        },
        "FA": {
            "feat_frac": [0.1, 0.5, 0.8, None]
        },
        "FastICA": {
            "feat_frac": [0.1, 0.5, 0.8, None],
            "algorithm": ["parallel", "deflation"]
        },
        "PCA": {
            "feat_frac": [None, 0.01, 0.1, 0.25, 0.5, 0.75, 1],
            "n_components": ["mle", 2, 3],
            "whiten": [False, True],
        },
        # "KernelPCA": {
        #    "feat_frac": [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
        #    "kernel": ["linear", "poly", "rbf", "sigmoid", "cosine"]
        # },
        "kBest": {
            "feat_frac": [0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 1],
            "score_func": [feature_selection.chi2, feature_selection.f_classif]
        },
        "MaxAbsScaler": {},
        "MinMaxScaler": {},
        "Normalizer": {},
        "StandardScaler": {}
    }

    kwargs_dict = {**ensemble_kwargs, **clf_kwargs, **transform_kwargs}
    return kwargs_dict



# -*- coding: utf-8 -*-

"""
This module defines the structure of GP primitives.

The GP primitives are nodes with typed edges (parent input and child output types must match) and variable
arity (for a given type, its final arity can be chosen during the evolution process).

A ``GpPrimitive`` is a node whose types, arities and keyword arguments have been decided. To create such primitives,
it is possible to take use of templates. These contain possible values of arities, types and keyword arguments and
methods for choosing final values.

The primitive templates defined in this file are
    1) functions - inner nodes of the tree, transform input into output
    2) terminals - leaves of the tree, provide constant output.
"""


import functools
import random

from copy import deepcopy
from deap import base
from typing import Callable, List, Any, Dict, Union, Tuple


class GpTreeIndividual:
    """Represents a tree individual used in GP.
    The individual is a tree encoded as a list of ``GpPrimitive`` nodes.

    The list is a post-order representation of the tree. The tree can be
    uniquely reconstructed using the arity (and types) of primitives.
    """
    def __init__(self, prim_list: List['GpPrimitive'], max_height: int):
        """
        Construct a GP tree from a list of primitives.

        Args:
            prim_list: list of GpPrimitive prim_list: Post-order representation of the tree.
            max_height: Height of the tree - maximum of all node depths + 1
        """
        self.primitives = prim_list
        self.max_height = max_height

        self.validate_tree()

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))

        return new

    def __eq__(self, other):
        if not isinstance(other, GpTreeIndividual):
            return False

        if self.primitives != other.primitives:
            return False

        return True

    def __repr__(self):
        return f'GpTreeIndividual(height={self.max_height} primitives={self.primitives.__repr__()})'

    def run_tree(self, node_func, group_children: bool = False) -> Any:
        """
        Applies a function with the signature ``func(node, child_list)`` on all
        nodes in the tree. The tree is traversed in post-order.

        The arguments of the function are a node and list of result values of its child nodes.

        :param node_func: Function which is applied on all nodes of the tree.
        :return: Return value of the root.
        """
        stack = []

        for node in self.primitives:
            if node.arity == 0:
                stack.append(node_func(node, []))
            else:
                args = stack[-node.arity:]
                stack = stack[:-node.arity]

                if group_children:
                    children_by_type = []
                    for t in node.in_type:
                        t_children, args = args[-t.arity:], args[:-t.arity]
                        children_by_type.append((t.name, t_children))

                    args = children_by_type

                stack.append(node_func(node, args))

        if len(stack) > 1:
            raise ValueError("Bad tree - invalid primitive list.")

        return stack.pop()

    def subtree(self, root_ind: int) -> Tuple[int, int]:
        """
        Returns the start position of the subtree with primitive `self.primitives[root_ind]` as root. Note
        that the returned value is in fact the index of one of the leaves, as the node list is post-order.
        As so, the whole subtree is extracted with `self.primitives[subtree(root_ind), root_ind + 1]`.

        Args:
            root_ind: Position of the root (index of the beginning).

        Returns: A tuple `(s, h)`, where `s` is the start index and `h` is the height of the subtree.
        """
        curr = self.primitives[root_ind]

        arity_rem = curr.arity
        init_h = curr.depth
        max_h = init_h

        while arity_rem > 0:
            root_ind = root_ind - 1
            curr = self.primitives[root_ind]
            max_h = max(max_h, curr.depth)

            arity_rem = arity_rem - 1 + curr.arity

        return root_ind, (max_h - init_h + 1)

    def validate_tree(self):
        """
        Validates the tree, raises an exception if its children are invalid.
        """

        def validate_node(node, child_list):
            if node.arity != len(child_list):
                raise ValueError("Invalid number of children.")

            child_id = 0
            for in_type in node.in_type:
                for i in range(in_type.arity):
                    child = child_list[child_id + i]
                    if child.out_type != in_type.name:
                        raise ValueError("Invalid child type.")

                    if child.depth != node.depth + 1:
                        raise ValueError("Invalid child height.")

                child_id += in_type.arity

            return node

        self.run_tree(validate_node)

        if self.max_height != max(prim.depth + 1 for prim in self.primitives):
            raise ValueError("Invalid tree height.")


class DeapTreeIndividual(GpTreeIndividual):
    """
    Represents an individual with DEAP-compatible fitness.
    """
    def __init__(self, prim_list: List['GpPrimitive'], max_height: int):
        super().__init__(prim_list, max_height)

        self.fitness = DeapTreeIndividual.Fitness()  # (score, log(evaluation_time))
        self.compiled_pipe = None

    class Fitness(base.Fitness):
        def __init__(self, values=()):
            self.weights = (1.0, -1.0)
            super().__init__(values)

    def reset(self):
        del self.fitness.values

        self.compiled_pipe = None


class GpPrimitive:
    """
    Represents a typed node in the GP tree.

    Its name and keyword dictionary hold information about the function
    or object, which is associated with the node.
    """

    def __init__(self,
                 name: str,
                 obj_kwargs: Dict[str, Any],
                 in_type: List['GpPrimitive.InputType'],
                 out_type: str,
                 arity: int,
                 depth: int):
        """
        Creates an instance of a GP tree node. The number and output types as well as the ordering of its children
        is specified by `in_type`.

        Args:
            name: Name of the node.
            obj_kwargs: Keyword arguments associated with the node.
            in_type:
                List of input types with arity. The subtypes are ordered - e.g. [('data', 2), ('ens', 1)] is not the
                same as [('ens', 1), ('data', 2)].
            arity: Sum of arity of subtypes.
            depth: Depth of the node.
        """
        self.name = name
        self.obj_kwargs = obj_kwargs
        self.in_type = in_type
        self.out_type = out_type
        self.arity = arity
        self.depth = depth

    def __deepcopy__(self, memo=None):
        if memo is None:
            memo = {}

        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))

        return new

    def __eq__(self, other):
        if not isinstance(other, GpPrimitive):
            return False

        if self.name != other.name:
            return False

        if self.arity != other.arity or self.in_type != other.in_type or self.out_type != other.out_type:
            return False

        if self.obj_kwargs != other.obj_kwargs:
            return False

        return True

    def __repr__(self):
        return 'GpPrimitive(name=' + self.name + ", arity={}".format(self.arity)\
               + ", height={})".format(self.depth)

    class InputType:
        """
        Represents the input type of a primitive. It determines how many children with a specific output type
        should the node have.
        """
        def __init__(self, name: str, arity: int):
            """
            Construct a new instance of input type with arity.

            :param name: Name of the type.
            :param arity: Arity of this type.
            """
            self.name = name
            self.arity = arity

        def __eq__(self, other):
            if not isinstance(other, GpPrimitive.InputType):
                return False

            if self.name != other.name or self.arity != other.arity:
                return False

            return True


class GpTerminalTemplate:
    """
    Represents a terminal of the GP tree, or a primitive with no inputs.
    The output type is fixed.

    The keyword arguments are chosen from lists of possible values.
    """
    def __init__(self, name: str, out_type: str, group: str = None):
        """
        Creates a new instance of a terminal template.

        Args:
            name: Name of the node.
            out_type: Name of the output type.
        """
        self.name = name
        self.type_arity_template = []
        self.out_type = out_type
        self.group = group

    def __repr__(self):
        return f"GpTerminalTemplate: {self.name} - {self.group}"

    def create_primitive(self, curr_height: int, max_arity: int, kwargs_dict: Dict[str, List[Any]]) -> GpPrimitive:
        """
        Creates an instance of a `GpPrimitive` from the template.

        Selects keyword arguments from `kwargs_dict`. For every key,
        the dict contains a list of possible values.

        Args:
            curr_height: Height at which the node is generated.
            max_arity: Only for compatibility, not used for terminals.
            kwargs_dict: Dictionary which contains possible keyword argument values.

        Return: A new instance of `GpPrimitive`.
        """
        prim_kwargs = _choose_kwargs(kwargs_dict)

        return GpPrimitive(self.name, prim_kwargs, [], self.out_type, 0, curr_height)


class TypeArity:
    """
    Represents a variable node arity associated with a type.
    """
    def __init__(self,
                 prim_type: str,
                 arity_template: Union[int, Tuple[int, int], Tuple[int, str]]):
        """
        Constructs a new instance of a type template - with a fixed type, but possibly variable arity.
        Args:
            prim_type: Name of the type.
            arity_template: Either a fixed arity value, or a bounded interval (a, b) where a and b are integers,
                or an interval (a, 'n') that has only a lower bound a ('n' is a string).
        """
        self.prim_type = prim_type
        self.arity_template = arity_template

        # check arity range
        if isinstance(self.arity_template, tuple):
            lower_invalid = self.arity_template[0] < 0
            upper_invalid = self.arity_template[1] != 'n' and self.arity_template[0] > self.arity_template[1]

            if lower_invalid or upper_invalid:
                raise ValueError("Invalid arity range.")

        # check fixed arity
        elif isinstance(self.arity_template, int):
            if self.arity_template <= 0:
                raise ValueError("Arity must be greater than 0.")
        else:
            raise ValueError("Invalid arity type.")

    def is_valid_arity(self, arity: int):
        """
        Determines whether `self.choose_arity` could possibly result in `arity`.
        Args:
            arity: Input arity to compare with this template.

        Returns: True if `arity` can be created from this template.

        """
        if isinstance(self.arity_template, tuple):
            # out of range
            if arity < self.arity_template[0]:
                return False

            if self.arity_template[1] != 'n' and arity > self.arity_template[1]:
                return False

            # inside range
            return True

        # match fixed arity
        if isinstance(self.arity_template, int):
            return self.arity_template == arity

        return False

    def choose_arity(self, max_arity: int) -> int:
        """
        Chooses an integer arity from the arity range or
        returns the arity if it is already a fixed value.

        Args:
            max_arity: The upper bound of arities. It is an upper bound for all ranges, even when the arity
                upper bound is greater. If the lower bound is greater than this value, `max_arity` is not applied
                and the value is chosen from the original interval.

        Returns: A fixed arity value.
        """
        if not isinstance(self.arity_template, tuple):
            return self.arity_template

        a_from, a_to = self.arity_template

        # is not applied for ranges which are greater than ``max_arity`` as it could result in invalid behavior
        if a_to == 'n' or a_to > max_arity >= a_from:
            a_to = max(max_arity, a_from)

        return random.randint(a_from, a_to)
        

class GpFunctionTemplate:
    """
    Represents an inner node of the GP tree.
    This class is a template of a function node - the input type may have variable arity which is decided when
    an instance of `GpPrimitive` is created from this template. During this process, keyword arguments are decided
    as well.

    The template input type is an ordered list of subtypes. These may have a variable arity (a range) and the
    final arity is chosen during the instantiation.

    The keyword arguments are chosen from lists of possible values.
    """
    def __init__(self, name: str, type_arity_template: List[TypeArity], out_type: str, group: str = None):
        """

        Args:
            name: Name key associated with the node.
            type_arity_template: Ordered list of children subtypes with variable arity.
            out_type: Name of the output type.
            group: Name of group of the node.
        """

        self.name = name
        self.type_arity_template = type_arity_template
        self.out_type = out_type
        self.group = group

    def __repr__(self):
        return f"GpFunctionTemplate: {self.name} - {self.group}"

    def create_primitive(self,
                         curr_height: int,
                         max_arity: int,
                         kwargs_dict: Dict[str, List],
                         in_type: GpPrimitive.InputType = None) -> GpPrimitive:
        """
        Creates an instance of a ``GpPrimitive`` from the template.

        Selects keyword arguments from ``kwargs_dict``. For every key,
        the dict contains a list of possible values.

        Select final arities of every TypeArity in `self.type_arities` - that is, if a function node has
        a variable number of input arguments (possibly of different types), for every subtype a fixed arity
        is chosen.

        TODO add example

        Args:
            curr_height: Height at which the node is generated
            max_arity: Maximum arity value which can be chosen for a single TypeArity.
            kwargs_dict: Dictionary which contains possible keyword argument values.
            in_type: Input type; if provided, it is used instead of generating a new random type.

        Returns: A new instance of GpPrimitive
        """
        prim_kwargs = _choose_kwargs(kwargs_dict)

        def create_type(t):
            arity = t.choose_arity(max_arity)

            # do not construct a type in case of 0
            if arity == 0:
                return None

            return GpPrimitive.InputType(t.prim_type, arity)

        if in_type is None:
            # ordered list of final arity of types
            in_type = [create_type(t_a) for t_a in self.type_arity_template]
            in_type = [t for t in in_type if t is not None]

        #
        # total arity of the node
        arity_sum = functools.reduce(lambda s, t: s + t.arity, in_type, 0)
        
        return GpPrimitive(self.name, prim_kwargs, in_type, self.out_type, arity_sum, curr_height)


def _choose_kwargs(kwargs_dict: Dict[str, List]) -> Dict[str, Any]:
    """
    Chooses keyword argument values from the argument dictionary `kwargs_dict`.
    For every keyword argument, it contains a list of possible values for every key.

    Args:
        kwargs_dict: Dict of possible kwarg values.

    Returns: Dict with one value selected per key.

    """
    if kwargs_dict is None:
        return {}

    return {k: random.choice(arg_list) for k, arg_list in kwargs_dict.items()}


def tree_str(gp_tree: GpTreeIndividual, with_hyperparams=False):

    def get_tree(node, ch_l):
        return node, ch_l

    def print_node(node: GpPrimitive, ch_l, indent, str_res):
        str_res += "  " * indent + node.name + "\n"
        if with_hyperparams:
            for k, val in node.obj_kwargs.items():
                str_res += "  " * (indent + 1) + "| " + f"{k}: {val}\n"

        for child, next_list in ch_l:
            str_res = print_node(child, next_list, indent + 1, str_res)

        return str_res

    root, child_list = gp_tree.run_tree(get_tree)
    return print_node(root, child_list, 0, "")


import json
import warnings
import yaml

from functools import partial
from genens.config.utils import import_custom_func
from genens.gp.types import TypeArity, GpTerminalTemplate, GpFunctionTemplate

from typing import Union, Dict


class GenensConfig:
    """
    Configuration of Genens estimators. Contains settings of GP algorithm and individuals
    and configuration of methods that decode nodes into scikit-learn models.
    """
    def __init__(self, function_config=None, full_config=None, term_config=None, kwargs_config=None,
                 cx_pb=0.5, mut_pb=0.3, mut_args_pb=0.9, mut_node_pb=0.9,
                 hc_repeat=0, hc_keep_last=False, hc_mut_pb=0.2, hc_n_nodes=3,
                 mut_multiple_args=False, mut_multiple_nodes=False,
                 min_height=1, max_height=4, min_arity=2, max_arity=3, group_weights=None):
        """
        Creates a instance of genens configuration.

        Args:
            function_config: Function configuration dictionary, keys correspond do a particular primitive name.
            full_config: Node configuration dictionary, nodes are grouped by output types,
                contains node templates used during the grow phase.

            term_config: Node configuration dictionary, nodes are grouped by output types,
                contains node templates used to terminate a tree.

            kwargs_config: Keyword arguments for nodes, which are passed to functions during the decoding.
                For every argument, there is a list of possibles values to choose from during the evolution.

            min_height: Minimum height of trees.
            max_height: Maximum height of trees.
            min_arity: Lower arity limit on any subtype.
            max_arity: Upper arity limit on any subtype. A node may have a larger total arity than this
                limit, but none of its subtypes can exceed it.
                E.g. for `max_arity` == 2 the node type (out^2, data^1) is valid - the arity of the node
                is 3, but its subtypes do not exceed the limit.

            group_weights: Weights of node groups, used during node selection. Nodes with a higher weight
                have a greater probability to appear in trees.
        """
        self.func_config = function_config if function_config is not None else {}
        self.full_config = full_config if full_config is not None else {}
        self.term_config = term_config if term_config is not None else {}
        self.kwargs_config = kwargs_config if kwargs_config is not None else {}

        self.min_height = min_height
        self.max_height = max_height
        self.min_arity = min_arity
        self.max_arity = max_arity

        self.cx_pb = cx_pb
        self.mut_pb = mut_pb
        self.mut_args_pb = mut_args_pb
        self.mut_node_pb = mut_node_pb

        self.hc_repeat = hc_repeat
        self.hc_keep_last = hc_keep_last
        self.hc_mut_pb = hc_mut_pb
        self.hc_n_nodes = hc_n_nodes

        self.mut_multiple_nodes = mut_multiple_nodes
        self.mut_multiple_args = mut_multiple_args

        self.group_weights = group_weights if group_weights is not None else {}

    def add_terminal(self, prim: GpTerminalTemplate, leaf_only: bool = False):
        """
        Adds a new primitive to the configuration - to both full and term dictionaries.
        If `term_only` is True, it is added only to the terminal set.

        Args:
            prim: Primitive to be added.
            leaf_only: Specifies whether the primitive should be added only to the terminal set.
        """

        if not leaf_only:
            out_list = self.full_config.setdefault(prim.out_type, [])
            out_list.append(prim)

        out_list = self.term_config.setdefault(prim.out_type, [])
        out_list.append(prim)

    def add_function(self, prim):
        """
        Add a function to the configuration - to the full set.

        Args:
            prim: Primitive to be added.
        """
        out_list = self.full_config.setdefault(prim.out_type, [])
        out_list.append(prim)


def parse_config(config_path: str, base_config: GenensConfig = None, evo_kwargs: dict = None):
    config = base_config if base_config is not None else GenensConfig()

    with open(config_path, 'r') as input_file:
        yaml_config = yaml.safe_load(input_file)

    min_height = yaml_config.get('min_height', config.min_height)
    max_height = yaml_config.get('max_height', config.max_height)
    min_arity = yaml_config.get('min_arity', config.min_arity)
    max_arity = yaml_config.get('max_arity', config.max_arity)

    group_weights = yaml_config.get('group_weights', {})

    load_kwargs_from_yaml = evo_kwargs is None
    evo_kwargs = {} if evo_kwargs is None else evo_kwargs

    primitives = yaml_config.get('primitives', None)
    func_dict, full_dict, term_dict, evo_kwargs_dict = parse_primitives(primitives,
                                                                        parse_evo_kwargs=load_kwargs_from_yaml)

    return GenensConfig(function_config={**config.func_config, **func_dict},
                        full_config={**config.full_config, **full_dict},
                        term_config={**config.term_config, **term_dict},
                        kwargs_config={**config.kwargs_config, **evo_kwargs_dict, **evo_kwargs},
                        min_height=min_height,
                        max_height=max_height,
                        min_arity=min_arity,
                        max_arity=max_arity,
                        group_weights={**config.group_weights, **group_weights})


def parse_primitives(primitives, parse_evo_kwargs=True):
    func_dict = {}
    full_dict = {}
    term_dict = {}

    evo_kwargs_dict = {}

    for prim_key, values in primitives.items():
        parsed_prim = parse_primitive(prim_key, values)

        # optionally parse kwargs that are changed during evolution
        if 'evo_kwargs' in values:
            if not parse_evo_kwargs:
                warnings.warn("The parameter evo_kwargs is present yaml config, but it won't be loaded, "
                              "possibly because a dict with hyperparameters has been provided.")
            else:
                evo_kwargs_dict[prim_key] = values.get('evo_kwargs', {})

        prim = parsed_prim["prim"]
        # add to primitive sets
        if 'grow' in parsed_prim["set"]:
            _add_to_primset(prim, full_dict)

        if 'terminal' in parsed_prim["set"]:
            _add_to_primset(prim, term_dict)
            if "term_prim" in parsed_prim:
                _add_to_primset(parsed_prim["term_prim"], term_dict)

        func_dict[prim.name] = parsed_prim["func"]

    return func_dict, full_dict, term_dict, evo_kwargs_dict


def _add_to_primset(prim: Union[GpFunctionTemplate, GpTerminalTemplate], prim_dict: Dict):
    type_list = prim_dict.setdefault(prim.out_type, [])
    type_list.append(prim)


def parse_primitive(prim_name, prim_data):
    in_type = _parse_in_type(prim_data['in']) if 'in' in prim_data else None
    out_type = prim_data['out']

    func = _parse_func(prim_data['function'])
    group = prim_data.get('group', None)

    if in_type is None:
        prim = GpTerminalTemplate(prim_name, out_type, group=group)
    else:
        prim = GpFunctionTemplate(prim_name, in_type, out_type, group=group)

    parsed_prim = {
        'prim': prim,
        'func': func,
        'set': prim_data['set'].replace(' ', '').split(',')
    }

    if 'terminal_out' in prim_data:
        if 'terminal' not in parsed_prim['set']:
            warnings.warn('Terminal out type should be specified only for terminals, '
                          f'skipping for primitive {prim_name}')
        else:
            term_prim = GpTerminalTemplate(prim_name, prim_data['terminal_out'], group=group)
            parsed_prim['term_prim'] = term_prim

    return parsed_prim


def _parse_func(func_data: Union[str, Dict]):
    if isinstance(func_data, str):
        func_path = func_data
        func_kwargs = None
    else:
        func_path = func_data.pop('func')
        func_kwargs = func_data

    func = import_custom_func(func_path)
    return partial(func, **func_kwargs) if func_kwargs is not None else func


def _parse_in_type(in_type):
    result_type = []

    for subtype in in_type:
        type_name = subtype['name']
        if 'arity' in subtype:
            ta = TypeArity(type_name, subtype['arity'])
        elif 'from' in subtype and 'to' in subtype:
            ta = TypeArity(type_name, (subtype['from'], subtype['to']))
        else:
            raise ValueError("Invalid arity in config file.")

        result_type.append(ta)

    return result_type
